"""Convenience methods for asynchronous HTTP requests."""


async def get(session, url, params=None):
  async with session.get(url, params=params) as response:
    return await response.text()


async def post(session, url, payload=None):
  async with session.post(url, data=payload) as response:
    return await response.text()



import re
import sys
import logging
import asyncio
import textwrap

from hypotonic import request
from hypotonic.context import HtmlContext

logger = logging.getLogger('hypotonic')


async def get(session, _, data, url, params=None):
  response = await request.get(session, url, params)
  yield HtmlContext(url, response), data


async def post(session, _, data, url, payload=None):
  response = await request.post(session, url, payload)
  yield HtmlContext(url, response), data


async def find(_, context, data, selector):
  for result in context.select(selector):
    yield result, data


async def set(_, context, data, descriptor):
  if isinstance(descriptor, str):
    data = {**data, descriptor: context.text()}
  else:
    for key, selector in descriptor.items():
      if isinstance(selector, str):
        results = context.select(selector)
        data = {**data, key: results[0].text()}
      elif isinstance(selector, list):
        results = context.select(selector[0])
        values = [result.text() for result in results]
        data = {**data, key: values}
  yield context, data


async def follow(session, context, data, selector):
  for result in context.select(selector):
    url = result.text()
    response = await request.get(session, url)
    yield HtmlContext(url, response), data


async def paginate(session, context, data, selector, limit=sys.maxsize):
  while True:
    yield context, data
    limit -= 1
    results = context.select(selector)
    if limit <= 0 or len(results) == 0:
      break
    url = results[0].text()
    response = await request.get(session, url)
    context = HtmlContext(url, response)


async def filter(_, context, data, selector):
  if len(context.select(selector)) > 0:
    yield context, data


async def match(_, context, data, regex, flags=0):
  if re.search(regex, context.text(), flags=flags):
    yield context, data


async def delay(_, context, data, secs):
  await asyncio.sleep(secs)
  yield context, data


async def log(_, context, data, level=logging.INFO):
  logger.log(level, f"{len(data)}:{textwrap.shorten(str(context), width=72)}")
  yield context, data



import sys
import logging
import itertools
import importlib
import traceback

import asyncio
import aiohttp

logger = logging.getLogger('hypotonic')


class Hypotonic:
  def __init__(self, url=None):
    self.commands = []
    self.results = []
    self.errors = []

    if url:
      self.commands.append(('get', (url,), {}))

  async def worker(self, i, session, queue):
    logger.debug(f"Worker {i} starting.")

    module = importlib.import_module('hypotonic.command')
    while True:
      commands, context, data = await queue.get()

      try:
        command, args, kwargs = next(commands)
        logger.debug(("Start", i, command, args, kwargs))

        # Dynamically load the command function.
        func = getattr(module, command)
        async for result in func(session, context, data, *args, **kwargs):
          # Tee creates a shallow copy, advanced separately from original list.
          _, commands_copy = itertools.tee(commands)
          queue.put_nowait((commands_copy, *result))

        logger.debug(("Stop", i, command, args, kwargs))
      except StopIteration:
        self.results.append(data)
      except:
        self.errors.append(((command, args, kwargs, context), sys.exc_info()))
        logger.info(traceback.format_exc())
      finally:
        queue.task_done()

  async def run(self):
    session = aiohttp.ClientSession()
    queue = asyncio.Queue()

    tasks = []
    for i in range(4):
      loop = asyncio.get_event_loop()
      tasks.append(loop.create_task(self.worker(i, session, queue)))

    queue.put_nowait((iter(self.commands), None, {}))
    await queue.join()

    for task in tasks:
      task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)

    await session.close()

  def data(self):
    """Return all the scraped data as a list of dicts."""
    loop = asyncio.new_event_loop()
    try:
      loop.run_until_complete(self.run())
      return self.results, self.errors
    finally:
      loop.close()

  def __getattr__(self, attr):
    def apply(*args, **kwargs):
      self.commands.append((attr, args, kwargs))
      return self

    return apply



import unicodedata
from lxml import html
from parsel import csstranslator
from cssselect import SelectorError


class StringContext:
  def __init__(self, url, str):
    self.url = url
    self.str = str

  def select(self, selector):
    raise TypeError('Cannot run selector on a StringContext.')

  def text(self):
    return self.str

  def __str__(self):
    return self.str


class HtmlContext:
  def __init__(self, url, str_or_element):
    self.url = url
    self.element = str_or_element

    if isinstance(str_or_element, str):
      self.element = self.parse(self.url, self.element)

  def select(self, selector):
    selected = []
    for element in self.element.xpath(self.to_xpath(selector)):
      # XPath query can return both element, or lxml.etree._ElementUnicodeResult.
      if isinstance(element, str):
        selected.append(StringContext(self.url, element))
      else:
        selected.append(HtmlContext(self.url, element))
    return selected

  def text(self):
    """Convenience method to extract text from HTML element tree. Performs
      stripping and canonicalization of UTF-8 characters (e.g. '/xa0' to ' ')."""
    return unicodedata.normalize('NFKC', self.element.text_content().strip())

  def __str__(self):
    """Convert HTML element to raw html string including element tags."""
    return html.tostring(self.element).decode('utf-8')

  @staticmethod
  def to_xpath(selector):
    """Attempt to convert CSS selector to XPath."""
    try:
      return csstranslator.css2xpath(selector)
    except SelectorError:
      return selector

  @staticmethod
  def parse(url, html_string):
    doc = html.fromstring(html_string)
    # Making links absolute is required to allow following.
    doc.make_links_absolute(url)
    # Replacing <br> tags with \n, prevents text contatenating.
    for br in doc.xpath('*//br'):
      br.tail = '\n' + br.tail if br.tail else '\n'
    return doc



from setuptools import setup, find_packages

setup(name='hypotonic',
      version='0.0.7',
      description='Fast asynchronous web scraper with minimalist API.',
      long_description='Fast asynchronous web scraper with minimalist API.',
      classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Text Processing :: Markup :: HTML'
      ],
      url='https://github.com/mscavnicky/hypotonic',
      author='Martin Scavnicky',
      author_email='martin.scavnicky@gmail.com',
      license='MIT',
      packages=find_packages(exclude=['tests']),
      install_requires=['aiohttp', 'parsel', 'lxml', 'cssselect'],
      include_package_data=True,
      zip_safe=False)
